# The file is used to list the public API for document generation purpose.
from tensorflow_similarity import architectures
from tensorflow_similarity import evaluators
from tensorflow_similarity import layers
from tensorflow_similarity import losses
from tensorflow_similarity import metrics
from tensorflow_similarity import matchers
from tensorflow_similarity import models
from tensorflow_similarity import samplers
from tensorflow_similarity import tables
from tensorflow_similarity import visualization
# from tensorflow_similarity import algebra  # not public
from tensorflow_similarity import callbacks
from tensorflow_similarity import distance_metrics
from tensorflow_similarity import distances
from tensorflow_similarity import indexer
from tensorflow_similarity import metrics
# from tensorflow_similarity import types  # not public
from tensorflow_similarity import utils
