"""
Contrastive learning specialized losses.
"""
from .pn_loss import PNLoss  # noqa
from .triplet_loss import TripletLoss  # noqa
from .metric_loss import MetricLoss  # noqa
from .circle_loss import CircleLoss  # noqa
from .multisim_loss import MultiSimilarityLoss  # noqa
